from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageValue

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class Frostgnaw(ElementalSkillBase):
    name: Literal["Frostgnaw"] = "Frostgnaw"
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = ElementalSkillBase.get_cost(ElementType.CRYO)


class GlacialWaltz(ElementalBurstBase):
    name: Literal["Glacial Waltz"] = "Glacial Waltz"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_team_status("Icicle")]


# Talents


class ColdBloodedStrike_3_3(SkillTalent):
    name: Literal["Cold-Blooded Strike"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Kaeya"] = "Kaeya"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO,
        elemental_dice_number=4,
    )
    skill: Literal["Frostgnaw"] = "Frostgnaw"
    usage: int = 1

    def equip(self, match: Any) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = 1
        return super().equip(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset Usage
        """
        self.usage = 1
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached character use elemental skill, then heal itself.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not same player or character, or not skill, or not equipped
            return []
        if self.usage <= 0:
            # out of usage
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        skill = character.get_object(event.action.position)
        assert skill is not None
        if skill.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        # heal itself
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageElementalType.HEAL,
                        target_position=character.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            )
        ]


# character base


class Kaeya_3_3(CharacterBase):
    name: Literal["Kaeya"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | Frostgnaw | GlacialWaltz] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Ceremonial Bladework",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Frostgnaw(),
            GlacialWaltz(),
        ]


register_class(Kaeya_3_3 | ColdBloodedStrike_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost

from ...action import MakeDamageAction

from ...event import SkillEndEventArguments

from ...consts import DamageElementalType, DamageType, SkillType

from ...modifiable_values import DamageDecreaseValue, DamageIncreaseValue, DamageValue
from .base import DefendTeamStatus, ShieldTeamStatus


class FullPlate_3_3(ShieldTeamStatus):
    name: Literal["Full Plate"] = "Full Plate"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2

    def value_modifier_DAMAGE_MULTIPLY(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        """
        If this shield exists, decrease physical damage by half.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_receive_damage(
            self.position,
            match,
        ):
            # not this character receive damage, not modify
            return value
        if value.damage_elemental_type != DamageElementalType.PHYSICAL:
            # not physical damage, not modify
            return value
        assert self.usage > 0
        value.damage = (value.damage + 1) // 2  # round up
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        if skill is a normal attack, and is used by our Noelle, and she has
        talent I Got Your Back, heal all our characters for 1 HP.

        TODO: from description, HP heal is triggered by the shield, not talent.
        The difference is when Noelle uses normal attack, other team status
        (e.g. Sparks 'n' Splash) triggers before the shield and removes this
        shield, whether HP heal will be triggered.
        """
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not using normal attack
            return []
        if self.position.player_idx != event.action.position.player_idx:
            # not this player using skill
            return []
        table = match.player_tables[self.position.player_idx]
        character = table.get_active_character()
        if character.name != "Noelle":
            # not Noelle
            return []
        if character.talent is None or character.talent.name != "I Got Your Back":
            # not Noelle with talent I Got Your Back
            return []
        # heal all characters for 1 HP.
        ret: List[MakeDamageAction] = []
        for character in table.characters:
            if character.is_alive:
                ret.append(
                    MakeDamageAction(
                        damage_value_list=[
                            DamageValue(
                                position=self.position,
                                damage_type=DamageType.HEAL,
                                target_position=character.position,
                                damage=-1,
                                damage_elemental_type=DamageElementalType.HEAL,
                                cost=Cost(),
                            )
                        ],
                    )
                )
        return ret


class JadeScreen_3_3(DefendTeamStatus):
    name: Literal["Jade Screen"] = "Jade Screen"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 2
    max_in_one_time: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If our Ningguang has talent, then increase our geo damage by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not our player
            return value
        if value.damage_elemental_type != DamageElementalType.GEO:
            # not geo damage
            return value
        characters = match.player_tables[self.position.player_idx].characters
        ningguang_talent = False
        for character in characters:
            if character.name == "Ningguang" and character.talent is not None:
                ningguang_talent = True
                break
        if not ningguang_talent:
            # no ningguang with talent
            return value
        # increase damage
        value.damage += 1
        return value


class JadeShield_3_7(ShieldTeamStatus):
    name: Literal["Jade Shield"] = "Jade Shield"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2


register_class(FullPlate_3_3 | JadeScreen_3_3 | JadeShield_3_7)



"""
Base class of characters. Each character should have its own file, the file
contains the character class, talent class, talent card class, skill classes.
DO NOT implement status, summons, weapons, artifacts, etc. in this file, which
will break the import loop.
"""


from typing import List, Literal, Any, Tuple
from pydantic import validator

from ...utils.class_registry import register_base_class
from ...utils import accept_same_or_higher_version, get_instance
from ..event import (
    CharacterReviveEventArguments,
    GameStartEventArguments,
    MoveObjectEventArguments,
    UseSkillEventArguments,
)
from ..consts import (
    ELEMENT_TO_DIE_COLOR,
    DamageElementalType,
    DamageType,
    ObjectType,
    PlayerActionLabels,
    SkillType,
    WeaponType,
    ElementType,
    FactionType,
    ObjectPositionType,
    CostLabels,
)
from ..object_base import ObjectBase, CardBase
from ..struct import Cost, DeckRestriction, ObjectPosition
from ..modifiable_values import DamageValue
from ..action import (
    ChargeAction,
    CreateObjectAction,
    MakeDamageAction,
    MoveObjectAction,
    RemoveObjectAction,
    Actions,
    SkillEndAction,
    UseSkillAction,
)
from ..status import CharacterStatusBase
from ..card import ArtifactBase
from ..card import WeaponBase


class SkillBase(ObjectBase):
    """
    Base class of skills.
    """

    name: str
    type: Literal[ObjectType.SKILL] = ObjectType.SKILL
    skill_type: SkillType
    damage_type: DamageElementalType
    damage: int
    cost: Cost
    cost_label: int
    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # set cost label into cost
        self.cost.label = self.cost_label
        # based on character id, calculate skill id.

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        Get the action type of using the skill.

        Returns:
            Tuple[int, bool]: The first element is the action label, the second
                element is whether the action type is a combat action.
        """
        return PlayerActionLabels.SKILL.value, True

    def is_valid(self, match: Any) -> bool:
        """
        Check if the skill can be used.
        """
        return True

    def get_actions(self, match: Any) -> List[Actions]:
        """
        The skill is triggered, and get actions of the skill.
        By default, it will generate three actions:
        1. MakeDamageAction to attack the enemy active character with damage
            `self.damage` and damage type `self.damage_type`.
        2. ChargeAction to charge the active character by 1.
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Any
    ):  # pragma: no cover
        """
        When use skill event triggered, check if this skill is used.
        If so, return self.get_actions()
        """
        if event.action.skill_position == self.position:
            ret = self.get_actions(match)
            return ret
        return []

    # commonly used function for skills

    def is_talent_equipped(self, match: Any):
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return character.talent is not None

    def create_team_status(self, name: str, args: Any = {}) -> CreateObjectAction:
        position = ObjectPosition(
            player_idx=self.position.player_idx,
            area=ObjectPositionType.TEAM_STATUS,
            id=-1,
        )
        return CreateObjectAction(
            object_name=name, object_position=position, object_arguments=args
        )

    def create_character_status(self, name: str, args: Any = {}) -> CreateObjectAction:
        return CreateObjectAction(
            object_name=name,
            object_position=self.position.set_area(ObjectPositionType.CHARACTER_STATUS),
            object_arguments=args,
        )

    def create_opposite_character_status(
        self, match: Any, name: str, args: Any = {}
    ) -> CreateObjectAction:
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return CreateObjectAction(
            object_name=name,
            object_position=target.position.set_area(
                ObjectPositionType.CHARACTER_STATUS
            ),
            object_arguments=args,
        )

    def create_summon(self, name: str, args: Any = {}) -> CreateObjectAction:
        position = ObjectPosition(
            player_idx=self.position.player_idx, area=ObjectPositionType.SUMMON, id=-1
        )
        return CreateObjectAction(
            object_name=name, object_position=position, object_arguments=args
        )

    def charge_self(self, charge: int) -> ChargeAction:
        return ChargeAction(
            player_idx=self.position.player_idx,
            character_idx=self.position.character_idx,
            charge=charge,
        )

    def attack_opposite_active(
        self,
        match: Any,
        damage: int,
        damage_type: DamageElementalType,
    ) -> MakeDamageAction:
        assert damage > 0
        target_table = match.player_tables[1 - self.position.player_idx]
        target_character_idx = target_table.active_character_idx
        target_character = target_table.characters[target_character_idx]
        return MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.DAMAGE,
                    target_position=target_character.position,
                    damage=damage,
                    damage_elemental_type=damage_type,
                    cost=self.cost.copy(),
                )
            ],
        )

    def attack_self(
        self,
        match: Any,
        damage: int,
        damage_elemental_type: DamageElementalType
        | None = DamageElementalType.PIERCING,
    ) -> MakeDamageAction:
        """
        Attack self with damage `damage` and type `damage_elemental_type`.
        If `damage` is negative, it will heal self, and `damage_elemental_type`
        is ignored. Cannot damage with 0 damage.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        damage_type = DamageType.DAMAGE
        assert damage != 0
        if damage < 0:
            damage_type = DamageType.HEAL
            damage_elemental_type = DamageElementalType.HEAL
        return MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=damage_type,
                    target_position=character.position,
                    damage=damage,
                    damage_elemental_type=damage_elemental_type,
                    cost=Cost(),
                )
            ],
        )

    def element_application_self(
        self, match: Any, element: DamageElementalType
    ) -> MakeDamageAction:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.ELEMENT_APPLICATION,
                    target_position=character.position,
                    damage=0,
                    damage_elemental_type=element,
                    cost=self.cost.copy(),
                )
            ],
        )


class PhysicalNormalAttackBase(SkillBase):
    """
    Base class of physical normal attacks.
    """

    skill_type: Literal[SkillType.NORMAL_ATTACK] = SkillType.NORMAL_ATTACK
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    damage: int = 2
    cost_label: int = CostLabels.NORMAL_ATTACK.value

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number=1,
            any_dice_number=2,
        )


class ElementalNormalAttackBase(SkillBase):
    """
    Base class of elemental normal attacks.
    """

    skill_type: Literal[SkillType.NORMAL_ATTACK] = SkillType.NORMAL_ATTACK
    damage_type: DamageElementalType
    damage: int = 1
    cost_label: int = CostLabels.NORMAL_ATTACK.value

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number=1,
            any_dice_number=2,
        )


class ElementalSkillBase(SkillBase):
    """
    Base class of elemental skills.
    """

    skill_type: Literal[SkillType.ELEMENTAL_SKILL] = SkillType.ELEMENTAL_SKILL
    damage_type: DamageElementalType
    damage: int = 3
    cost_label: int = CostLabels.ELEMENTAL_SKILL.value

    @staticmethod
    def get_cost(element: ElementType) -> Cost:
        return Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number=3,
        )


class ElementalBurstBase(SkillBase):
    """
    Base class of elemental bursts.
    """

    skill_type: Literal[SkillType.ELEMENTAL_BURST] = SkillType.ELEMENTAL_BURST
    damage_type: DamageElementalType
    cost_label: int = CostLabels.ELEMENTAL_BURST.value

    @staticmethod
    def get_cost(element: ElementType, number: int, charge: int) -> Cost:
        return Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[element],
            elemental_dice_number=number,
            charge=charge,
        )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        When using elemental burst, the charge of the active character will be
        reduced by `self.charge`.
        """
        return [
            self.charge_self(-self.cost.charge),
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]


class AOESkillBase(SkillBase):
    """
    Base class that deals AOE damage. Can inherit this class after previous
    classes to do AOE attack. It will attack active character damage+element,
    back character back_damage_piercing.
    """

    back_damage: int

    def attack_opposite_active(
        self,
        match: Any,
        damage: int,
        damage_type: DamageElementalType,
    ) -> MakeDamageAction:
        """
        For AOE, modify attack opposite to also attack back characters
        """
        target_table = match.player_tables[1 - self.position.player_idx]
        target_character_idx = target_table.active_character_idx
        target_character = target_table.characters[target_character_idx]
        characters = target_table.characters
        damage_action = MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.DAMAGE,
                    target_position=target_character.position,
                    damage=damage,
                    damage_elemental_type=damage_type,
                    cost=self.cost.copy(),
                )
            ],
        )
        for cid, character in enumerate(characters):
            if cid == target_table.active_character_idx:
                continue
            if character.is_defeated:
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.DAMAGE,
                    target_position=character.position,
                    damage=self.back_damage,
                    damage_elemental_type=DamageElementalType.PIERCING,
                    cost=self.cost.copy(),
                )
            )
        return damage_action


class PassiveSkillBase(SkillBase):
    """
    Base class of passive skills.
    It has no cost and is always invalid (cannot be used).
    It has triggers to make effects.
    """

    skill_type: Literal[SkillType.PASSIVE] = SkillType.PASSIVE
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    damage: int = 0
    cost: Cost = Cost()
    cost_label: int = 0

    def is_valid(self, match: Any) -> bool:
        """
        Passive skills are always invalid.
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Passive skills are always invalid, so it will return an empty list.
        """
        raise AssertionError("Try to get actions of a passive skill")
        return []


class CreateStatusPassiveSkill(PassiveSkillBase):
    """
    This passive skill will generate a character status when game start
    for the character.

    By default, the status will re-generate when the character revive.
    If `regenerate_when_revive` is set to False, the status will not
    re-generate when the character revive.
    """

    status_name: str
    regenerate_when_revive: bool = True

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain status
        """
        return [self.create_character_status(self.status_name)]

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When character revive, gain status
        """
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.character_idx != self.position.character_idx
        ):
            # not this character, do nothing
            return []
        if self.regenerate_when_revive:
            return [self.create_character_status(self.status_name)]
        return []


class TalentBase(CardBase):
    """
    Base class of talents. Note almost all talents are skills, and will receive
    cost decrease from other objects.
    """

    name: str
    character_name: str
    type: Literal[ObjectType.TALENT] = ObjectType.TALENT
    cost_label: int = CostLabels.CARD.value | CostLabels.TALENT.value
    remove_when_used: bool = False

    def get_deck_restriction(self) -> DeckRestriction:
        """
        For talent cards, should contain the corresponding character.
        """
        return DeckRestriction(type="CHARACTER", name=self.character_name, number=1)

    def is_valid(self, match: Any) -> bool:
        """
        Only corresponding character is active character can equip this card.
        """
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError("Talent is not in hand")
        table = match.player_tables[self.position.player_idx]
        character = table.characters[table.active_character_idx]
        return character.name == self.character_name and character.is_alive

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        For most talent cards, can quip only on active character.
        """
        table = match.player_tables[self.position.player_idx]
        character = table.characters[table.active_character_idx]
        return [character.position]

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        """
        Act the talent. will place it into talent area.
        When other talent is equipped, remove the old one.
        For subclasses, inherit this and add other actions (e.g. trigger
        corresponding skills)
        """
        assert target is not None
        ret: List[Actions] = []
        table = match.player_tables[target.player_idx]
        character = table.characters[target.character_idx]
        # check if need to remove current talent
        if character.talent is not None:
            ret.append(
                RemoveObjectAction(
                    object_position=character.talent.position,
                )
            )
        new_position = character.position.set_id(self.position.id)
        ret.append(
            MoveObjectAction(
                object_position=self.position, target_position=new_position
            )
        )
        return ret

    def equip(self, match: Any) -> List[Actions]:
        return []

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this talent is moved from hand to character, it is considered
        as equipped, and will call `self.equip`.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.HAND
            and event.action.target_position.area == ObjectPositionType.CHARACTER
        ):
            # this talent equipped from hand to character
            return self.equip(match)
        return []


class SkillTalent(TalentBase):
    """
    Talents that trigger skills. They will get skill as input, which is
    saved as a private variable. They are all equiment cards.
    """

    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    skill: str

    def is_valid(self, match: Any) -> bool:
        """
        Only corresponding character is active character can equip this card.
        """
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError("Talent is not in hand")
        table = match.player_tables[self.position.player_idx]
        character = table.characters[table.active_character_idx]
        return (
            character.name == self.character_name
            and character.is_alive
            and not character.is_stunned
        )

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        For skill talent, the action label contains SKILL and is a combat
        action.
        """
        return (PlayerActionLabels.CARD.value | PlayerActionLabels.SKILL.value, True)

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        ret = super().get_actions(target, match)
        assert len(ret) > 0
        assert ret[-1].type == "MOVE_OBJECT"
        target_position = ret[-1].target_position
        target_character = match.player_tables[target_position.player_idx].characters[
            target_position.character_idx
        ]
        skills: List[SkillBase] = target_character.skills
        for s in skills:
            if s.name == self.skill:
                skill = s
                break
        else:
            raise AssertionError("Skill not found")
        ret.append(UseSkillAction(skill_position=skill.position))
        # skill used, add SkillEndAction
        ret.append(
            SkillEndAction(
                position=skill.position,
                target_position=match.player_tables[1 - skill.position.player_idx]
                .get_active_character()
                .position,
                skill_type=skill.skill_type,
            )
        )
        return ret


class CharacterBase(ObjectBase):
    """
    Base class of characters.
    """

    name: str
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    type: Literal[ObjectType.CHARACTER] = ObjectType.CHARACTER
    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )

    element: ElementType
    max_hp: int
    max_charge: int
    skills: List[SkillBase]

    # labels
    faction: List[FactionType]
    weapon_type: WeaponType

    # character status
    hp: int = -1
    charge: int = 0
    element_application: List[ElementType] = []
    is_alive: bool = True

    # objects attached to the character, including equips and character status
    attaches: List[CharacterStatusBase | WeaponBase | ArtifactBase | TalentBase] = []

    @validator("attaches", each_item=True, pre=True)
    def parse_attaches(cls, v):
        obj = get_instance(
            CharacterStatusBase | WeaponBase | ArtifactBase | TalentBase, v
        )
        return obj

    @validator("version", pre=True)
    def accept_same_or_higher_version(cls, v: str, values):
        return accept_same_or_higher_version(cls, v, values)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.hp == -1:
            self.hp = self.max_hp
        old_skill_ids = [x.id for x in self.skills]
        self._init_skills()
        self.position = self.position  # set position to update skill positions
        if len(old_skill_ids):
            new_skill_ids = [x.id for x in self.skills]
            if len(new_skill_ids) != len(old_skill_ids):
                raise AssertionError("Skill number changed after init")
            for id, skill in zip(old_skill_ids, self.skills):
                skill.id = id
                skill.position = skill.position.set_id(id)

    def __setattr__(self, name: str, value: Any) -> None:
        """
        When position is edited, update skill positions.
        """
        super().__setattr__(name, value)
        if name == "position":
            if self.position.player_idx < 0:
                # invalid position, do not update skill positions
                return
            for skill in self.skills:
                skill.position = ObjectPosition(
                    player_idx=self.position.player_idx,
                    character_idx=self.position.character_idx,
                    area=ObjectPositionType.SKILL,
                    id=skill.id,
                )

    def _init_skills(self) -> None:
        """
        Initialize skills. It will be called in __init__.
        """
        raise NotImplementedError

    @property
    def status(self) -> List[CharacterStatusBase]:
        """
        Get all status of the character.
        """
        return [x for x in self.attaches if isinstance(x, CharacterStatusBase)]

    @property
    def weapon(self) -> WeaponBase | None:
        """
        Get weapon of the character.
        """
        for x in self.attaches:
            if isinstance(x, WeaponBase):
                return x
        return None

    @property
    def artifact(self) -> ArtifactBase | None:
        """
        Get artifact of the character.
        """
        for x in self.attaches:
            if isinstance(x, ArtifactBase):
                return x
        return None

    @property
    def talent(self) -> TalentBase | None:
        """
        Get talent of the character.
        """
        for x in self.attaches:
            if isinstance(x, TalentBase):
                return x
        return None

    def remove_equip(
        self, equip_type: ObjectType
    ) -> WeaponBase | ArtifactBase | TalentBase:
        """
        Remove equip from character, and return the removed object.
        """
        assert equip_type in [ObjectType.WEAPON, ObjectType.ARTIFACT, ObjectType.TALENT]
        for idx, obj in enumerate(self.attaches):
            if obj.type == equip_type:
                ret = self.attaches.pop(idx)
                assert isinstance(ret, WeaponBase | ArtifactBase | TalentBase)
                return ret
        raise AssertionError("Equip not found")

    @property
    def is_defeated(self) -> bool:
        return not self.is_alive

    @property
    def is_stunned(self) -> bool:
        """
        Check if the character is stunned.
        """
        stun_status_names = [
            "Frozen",
            "Petrification",
            "Mist Bubble",
            "Stun",
        ]
        for status in self.status:
            if status.name in stun_status_names:
                return True
        return False

    @property
    def damage_taken(self) -> int:
        """
        Get damage taken by the character.
        """
        return self.max_hp - self.hp

    def get_object_lists(self) -> List[ObjectBase]:
        """
        Get all objects of the character, order is passive skill, weapon,
        artifact, talent, status. For status, order is their index in status
        list, i.e. generated time.
        """
        result: List[ObjectBase] = [self]
        for skill in self.skills:
            result.append(skill)
        result += self.attaches
        return result

    def get_object(self, position: ObjectPosition) -> ObjectBase | None:
        """
        Get object by its position. If object not exist, return None.
        """
        if position.area == ObjectPositionType.CHARACTER_STATUS:
            for status in self.status:
                if status.id == position.id:
                    return status
            return None
        elif position.area == ObjectPositionType.SKILL:
            for skill in self.skills:
                if skill.id == position.id:
                    return skill
            raise AssertionError("Skill not found")
        else:
            assert position.area == ObjectPositionType.CHARACTER
            if self.id == position.id:
                return self
            elif self.talent is not None and self.talent.id == position.id:
                return self.talent
            elif self.weapon is not None and self.weapon.id == position.id:
                return self.weapon
            elif self.artifact is not None and self.artifact.id == position.id:
                return self.artifact
            return None


register_base_class(CharacterBase)
register_base_class(TalentBase)



from typing import List, Literal, Any

from ....utils.class_registry import register_class

from ...event import SkillEndEventArguments

from ...action import Actions, MakeDamageAction
from ...consts import (
    ElementType,
    FactionType,
    SkillType,
    WeaponType,
    DamageElementalType,
    DamageType,
    DieColor,
)
from ..character_base import (
    AOESkillBase,
    PhysicalNormalAttackBase,
    ElementalSkillBase,
    ElementalBurstBase,
    CharacterBase,
    SkillTalent,
)
from ...struct import Cost
from ...modifiable_values import DamageValue
from ...summon.base import AttackerSummonBase


class Nightrider(ElementalSkillBase):
    name: Literal["Nightrider"] = "Nightrider"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Oz")]


class MidnightPhantasmagoria(ElementalBurstBase, AOESkillBase):
    name: Literal["Midnight Phantasmagoria"] = "Midnight Phantasmagoria"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    back_damage: int = 2
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
        charge=3,
    )


class StellarPredator_3_3(SkillTalent):
    name: Literal["Stellar Predator"]
    character_name: Literal["Fischl"] = "Fischl"
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Nightrider"] = "Nightrider"


class Oz_3_3(AttackerSummonBase):
    name: Literal["Oz"]
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1
    renew_type: Literal["RESET_WITH_MAX"] = "RESET_WITH_MAX"

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If Fischl made normal attack and with talent, make 2 electro damage
        to front.
        """
        action = event.action
        if action.skill_type != SkillType.NORMAL_ATTACK:
            # not using normal attack
            return []
        if self.position.player_idx != action.position.player_idx:
            # not attack by self
            return []
        character = match.player_tables[action.position.player_idx].characters[
            action.position.character_idx
        ]
        if (
            character.talent is not None
            and character.name == "Fischl"
            and character.talent.name == "Stellar Predator"
        ):
            # match, decrease usage, attack.
            # after make damage, will trigger usage check, so no need to
            # add RemoveObjectAction here.
            assert self.usage > 0
            self.usage -= 1
            target_table = match.player_tables[1 - self.position.player_idx]
            target_character = target_table.characters[
                target_table.active_character_idx
            ]
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=target_character.position,
                            damage=2,
                            damage_elemental_type=self.damage_elemental_type,
                            cost=Cost(),
                        )
                    ],
                )
            ]
        return []


class Fischl_3_3(CharacterBase):
    name: Literal["Fischl"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | Nightrider | MidnightPhantasmagoria] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT,
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Bolts of Downfall",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Nightrider(),
            MidnightPhantasmagoria(),
        ]


register_class(Fischl_3_3 | Oz_3_3 | StellarPredator_3_3)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.elemental_reaction import check_elemental_reaction
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from lpsim.server.consts import (
    DamageElementalType,
    ElementType,
    ElementalReactionType,
)
from tests.utils_for_test import (
    get_test_id_from_command,
    set_16_omni,
    check_hp,
    check_name,
    make_respond,
)


def test_crystallize():
    """
    agent 0: 1 elec to p1c2, then nothing until r3, and c2 will be defeated,
        choose c0. in r3, 3 geo to p1c2
    agent 1: in r1, 1 elec to p1c2, 1 geo to p1c2, 1 elec to p1c2, 1 geo to
        p1c2, in r2, 1 geo to p1c2, 1 elec to p1c2, 1 geo to p1c2, p0c2
        defeated, switch to p1c2, end.
    result: agent 1 get two crystallize, and agent 0 attack 3 + 1 - 2 = 2hp.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            # 'reroll',
            "skill 0 omni omni omni",
            "end",
            "end",
            "choose 0",
            "skill 1 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            # 'reroll',
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "GeoMobMage",
                "element": "GEO",
            },
            {
                "name": "GeoMobMage",
                "element": "GEO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0
    assert len(agent_1.commands) == 0
    assert match.round_number == 4
    check_hp(match, [[10, 10, 0], [10, 10, 7]])

    assert match.state != MatchState.ERROR


def test_frozen():
    """
    agent 0 goes first, sw 2, 1 cryo to p1c2, sw 0, 1 hydro to p1c2, sw 1,
        2 physical to p1c2, sw2, end, sw0, sw2, (cannot skill), end,
        (can skill) 1 cryo to p0c2.
    agent 1 sw0, sw2, (cannot skill), end, (can skill) 3 cryo to p0c2, sw 0,
        1 hydro to p0c2, end.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "end",
            "sw_char 0 omni",
            "sw_char 2 omni",
            "end",
            "skill 0 omni omni omni",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "sw_char 0 omni",
            "sw_char 2 omni",
            "end",
            "skill 1 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "HydroMobMage",
                "element": "HYDRO",
            },
            {
                "name": "HydroMob",
                "element": "HYDRO",
            },
            {
                "name": "CryoMobMage",
                "element": "CRYO",
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            if len(agent_0.commands) == 2:
                # p0c2 should be frozen, and cannot use skill
                check_name("Frozen", match.player_tables[0].characters[2].status)
                check_name("UseSkillRequest", match.requests, exist=False)
            elif len(agent_0.commands) == 1:
                # now not frozen and can use skill
                check_name(
                    "Frozen", match.player_tables[0].characters[2].status, exist=False
                )
                check_name("UseSkillRequest", match.requests, exist=True)
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 5:
                # p1c2 should be frozen, and cannot use skill
                check_name("Frozen", match.player_tables[1].characters[2].status)
                check_name("UseSkillRequest", match.requests, exist=False)
            elif len(agent_1.commands) == 4:
                # now not frozen and can use skill
                check_name(
                    "Frozen", match.player_tables[1].characters[2].status, exist=False
                )
                check_name("UseSkillRequest", match.requests, exist=True)
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0
    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[10, 10, 5], [9, 10, 3]])

    assert match.state != MatchState.ERROR


def test_frozen_and_pyro():
    """
    agent 0 goes first, sw 2, 1 cryo to p1c2, sw 0, 1 hydro to p1c2,
        1 hydro to p1c2, sw 1, 1 pyro to p1c2, end, 1 pyro to p1c2, sw 2,
        1 cryo to p1c2, 1 cryo to p1c2, sw 0, 1 hydro to p1c2, sw 1, end,
        1 pyro to p1c2, end.
    agent 1 end, end, end.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "end",
            "skill 0 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "end",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "HydroMobMage",
                "element": "HYDRO",
            },
            {
                "name": "PyroMobMage",
                "element": "PYRO",
            },
            {
                "name": "CryoMobMage",
                "element": "CRYO",
                "hp": 99,
                "max_hp": 99,
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0
    assert len(agent_1.commands) == 0
    assert match.round_number == 4
    check_hp(match, [[10, 10, 99], [10, 10, 82]])
    assert match.player_tables[1].characters[2].element_application == ["PYRO"]

    assert match.state != MatchState.ERROR


def test_burning_flame():
    """
    agent 0 goes first, sw 2, 1 pyro to p1c2, sw 0, 1 dendro to p1c2, end,
        1 dendro to p1c2, end, 1 dendro to p1c2, 1 dendro to p1c2, sw 2,
        1 pyro to p1c2, sw 0, 1 dendro to p1c2, end, 1 dendro to p1c2,
        sw 2, 1 pyro to p1c2, end.
    agent 1 sw2, end, end, end, end.
    result: 18 damage and 2 burning flame.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "end",
            "end",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "PyroMobMage",
                "element": "PYRO",
                "hp": 99,
                "max_hp": 99,
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0
    assert len(agent_1.commands) == 0
    assert match.round_number == 5
    assert len(match.player_tables[0].summons) == 1
    assert match.player_tables[0].summons[0].name == "Burning Flame"
    assert match.player_tables[0].summons[0].usage == 1
    check_hp(match, [[10, 10, 99], [10, 10, 80]])

    assert match.state != MatchState.ERROR


def test_dendro_core_catalyzing_field():
    """
    agent 0 goes first, sw 2, 1 electro to p1c2, sw 0, 1 dendro to p1c2,
        1 dendro to p1c2, sw 1, 1 hydro to p1c2, sw 2, end, 1 electro to p1c2,
        sw 1, 1 hydro to p1c2, 1 hydro to p1c2, sw 0, 1 dendro to p1c2, end,
        1 dendro to p1c2, sw 1, 1 hydro to p1c2, sw 2, 1 electro to p1c2,
        1 electro to p1c2, end.
    agent 1 sw2, end, end, end.
    tested catalyzing field and dendro core will not disappear at round end,
    can trigger simultaneously, and max dendro core number is 1.
    result: 22 damage and no status.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "end",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "sw_char 1 omni",
            "skill 0 omni omni omni",
            "sw_char 2 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "end",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "HydroMobMage",
                "element": "HYDRO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
                "hp": 99,
                "max_hp": 99,
            },
        ],
        "cards": [
            {
                "name": "Strategize",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_0.commands) == 0
    assert len(agent_1.commands) == 0
    assert match.round_number == 4
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[10, 10, 99], [9, 9, 76]])

    assert match.state != MatchState.ERROR


def test_catalyzing_field_old():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 1 omni omni omni",
            "sw_char 0 omni",
            "skill 0 omni omni omni",
            "skill 0 omni omni omni",
            "skill 3 omni omni omni",
            "card 0 0 omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida*2
        character:Fischl
        Mondstadt Hash Brown*30
        """
    )
    deck.characters[0].hp = 99
    deck.characters[0].max_hp = 99
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.event_handlers[0].version = "3.3"  # type: ignore
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_swirl():
    """
    first: swirl pyro and hydro
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 1 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:HydroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 1
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[6, 6, 6], [10, 10, 10]])

    assert match.state != MatchState.ERROR


def test_swirl_2():
    """
    second: swirl electro and hydro
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 1 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:HydroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 1
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[4, 6, 6], [10, 10, 10]])

    assert match.state != MatchState.ERROR


def test_swirl_3():
    """
    third: swirl cryo and hydro, and swirl pyro. To apply four elements,
    after match start, modify p1c0 skill 1 element to pyro.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 1 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "end",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:CryoMobMage
        character:HydroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.player_tables[1].characters[0].skills[
        1
    ].damage_type = DamageElementalType.PYRO
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 1:
                table = match.player_tables[0]
                characters = table.characters
                assert characters[0].element_application == ["PYRO"]
                assert characters[1].element_application == []
                assert characters[2].element_application == []
                assert len(characters[0].status) == 0
                assert len(characters[1].status) == 1
                assert len(characters[2].status) == 1
                assert characters[1].status[0].name == "Frozen"
                assert characters[2].status[0].name == "Frozen"
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 2
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[2, 4, 4], [10, 10, 10]])
    assert match.player_tables[0].characters[0].element_application == []
    assert match.player_tables[0].characters[1].element_application == ["PYRO"]
    assert match.player_tables[0].characters[2].element_application == ["PYRO"]
    assert len(match.player_tables[0].characters[1].status) == 0
    assert len(match.player_tables[0].characters[2].status) == 0

    assert match.state != MatchState.ERROR


def test_swirl_4():
    """
    fourth: swirl defeated enemys should have no damage on it.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 2 0 1 2",
            "skill 0 0 1 2",
            "end",
            "sw_char 2 0",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:CryoMobMage
        character:HydroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 2
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[0, 8, 9], [10, 10, 10]])
    assert match.player_tables[0].characters[0].element_application == []
    assert match.player_tables[0].characters[1].element_application == []
    assert match.player_tables[0].characters[2].element_application == ["CRYO"]

    assert match.state != MatchState.ERROR


def test_swirl_with_catalyzing_field():
    """
    four: electro and dendro to generate catalyzing field, and swirl electro
        will not trigger catalyzing field.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "sw_char 0 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            # cannot swirl dendro
            "TEST 1 enemy dendro none none",
            "sw_char 1 1",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "end",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            # no elemental application, 2 catalyzing field, field not increase
            # back electro damage.
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:DendroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    for character, app in zip(
                        match.player_tables[0].characters, [["DENDRO"], [], []]
                    ):
                        assert character.element_application == app
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert len(agent_0.commands) == 0
    assert match.round_number == 2
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[6, 4, 7], [10, 10, 10]])
    assert match.player_tables[0].characters[0].element_application == []
    assert match.player_tables[0].characters[1].element_application == []
    assert match.player_tables[0].characters[2].element_application == []
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == "Catalyzing Field"
    assert match.player_tables[1].team_status[0].usage == 2

    assert match.state != MatchState.ERROR


def test_swirl_with_catalyzing_field_and_dendro_core():
    """
    five: generate catalyzing field and dendro core, and swirl electro
        will not trigger catalyzing field.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "end",
            "end",
            "sw_char 1 0",
            "end",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "end",
            "skill 0 0 1 2",
            "end",
            "sw_char 3 0",
            "skill 0 0 1 2",
            "TEST 1 elec none elec elec, 1 dendro core 1 field",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:HydroMobMage
        character:ElectroMobMage
        character:DendroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 4
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[2, 8, 9, 9], [10, 10, 10, 10]])
                    assert match.player_tables[0].characters[0].element_application == [
                        "ELECTRO"
                    ]
                    assert (
                        match.player_tables[0].characters[1].element_application == []
                    )
                    assert match.player_tables[0].characters[2].element_application == [
                        "ELECTRO"
                    ]
                    assert match.player_tables[0].characters[3].element_application == [
                        "ELECTRO"
                    ]
                    assert len(match.player_tables[1].team_status) == 2
                    assert (
                        match.player_tables[1].team_status[0].name == "Catalyzing Field"
                    )
                    assert match.player_tables[1].team_status[0].usage == 1
                    assert match.player_tables[1].team_status[1].name == "Dendro Core"
                    assert match.player_tables[1].team_status[1].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert len(agent_0.commands) == 0
    assert match.round_number == 4
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[2, 8, 9, 9], [10, 10, 10, 9]])
    assert match.player_tables[0].characters[0].element_application == ["ELECTRO"]
    assert match.player_tables[0].characters[1].element_application == []
    assert match.player_tables[0].characters[2].element_application == ["ELECTRO"]
    assert match.player_tables[0].characters[3].element_application == ["ELECTRO"]
    assert len(match.player_tables[1].team_status) == 2
    assert match.player_tables[1].team_status[0].name == "Catalyzing Field"
    assert match.player_tables[1].team_status[0].usage == 1
    assert match.player_tables[1].team_status[1].name == "Dendro Core"
    assert match.player_tables[1].team_status[1].usage == 1

    assert match.state != MatchState.ERROR


def test_swirl_with_catalyzing_field_and_dendro_core_old_version():
    """
    five: generate catalyzing field and dendro core, and swirl electro
        will not trigger catalyzing field.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "end",
            "end",
            "sw_char 1 0",
            "end",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "end",
            "skill 0 0 1 2",
            "end",
            "sw_char 3 0",
            "skill 0 0 1 2",
            "TEST 1 elec none elec elec, 1 dendro core 1 field",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    match.event_handlers[0].version = "3.3"  # type: ignore
    deck = Deck.from_str(
        """
        default_version:4.0
        character:HydroMobMage
        character:ElectroMobMage
        character:DendroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 4
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[2, 8, 9, 9], [10, 10, 10, 10]])
                    assert match.player_tables[0].characters[0].element_application == [
                        "ELECTRO"
                    ]
                    assert (
                        match.player_tables[0].characters[1].element_application == []
                    )
                    assert match.player_tables[0].characters[2].element_application == [
                        "ELECTRO"
                    ]
                    assert match.player_tables[0].characters[3].element_application == [
                        "ELECTRO"
                    ]
                    assert len(match.player_tables[1].team_status) == 2
                    assert (
                        match.player_tables[1].team_status[0].name == "Catalyzing Field"
                    )
                    assert match.player_tables[1].team_status[0].usage == 2
                    assert match.player_tables[1].team_status[1].name == "Dendro Core"
                    assert match.player_tables[1].team_status[1].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert len(agent_0.commands) == 0
    assert match.round_number == 4
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[2, 8, 9, 9], [10, 10, 10, 9]])
    assert match.player_tables[0].characters[0].element_application == ["ELECTRO"]
    assert match.player_tables[0].characters[1].element_application == []
    assert match.player_tables[0].characters[2].element_application == ["ELECTRO"]
    assert match.player_tables[0].characters[3].element_application == ["ELECTRO"]
    assert len(match.player_tables[1].team_status) == 2
    assert match.player_tables[1].team_status[0].name == "Catalyzing Field"
    assert match.player_tables[1].team_status[0].usage == 2
    assert match.player_tables[1].team_status[1].name == "Dendro Core"
    assert match.player_tables[1].team_status[1].usage == 1

    assert match.state != MatchState.ERROR


def test_overloaded():
    """
    test overloaded a loop, and single character overloaded, and overloaded
    kill not need to choose character.
    """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "end",
            "end",
            "end",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "TEST 2 active 1",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 3 active 2",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "TEST 1 active 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 2 active 1",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 3 active 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 1 active 0",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 1 active 0",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:ElectroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.player_tables[0].characters[0].hp = 90
    match.player_tables[0].characters[0].max_hp = 90
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    assert match.player_tables[0].active_character_idx == 0
                if test_id == 2:
                    assert match.player_tables[0].active_character_idx == 1
                if test_id == 3:
                    assert match.player_tables[0].active_character_idx == 2
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 5
    assert len(match.player_tables[0].team_status) == 0
    check_hp(match, [[76, 0, 0], [10, 10, 10]])

    assert match.state != MatchState.ERROR


def test_background_overloaded():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "end",
        ],
        only_use_command=True,
    )
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:ElectroMobMage
        character:AnemoMobMage
        Prophecy of Submersion*10
        Stellar Predator*10
        Wine-Stained Tricorne*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.player_tables[0].characters[0].hp = 90
    match.player_tables[0].characters[0].max_hp = 90
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 2
    assert len(match.player_tables[0].team_status) == 0
    assert match.player_tables[0].active_character_idx == 0
    check_hp(match, [[86, 6, 6], [10, 10, 10]])

    assert match.state != MatchState.ERROR


def test_pure_elemental_reaction():
    c = ElementType.CRYO
    p = ElementType.PYRO
    h = ElementType.HYDRO
    e = ElementType.ELECTRO
    d = ElementType.DENDRO
    a = ElementType.ANEMO
    g = ElementType.GEO
    pc = ElementalReactionType.MELT
    pe = ElementalReactionType.OVERLOADED
    ec = ElementalReactionType.SUPERCONDUCT
    ch = ElementalReactionType.FROZEN
    aa = ElementalReactionType.SWIRL
    gg = ElementalReactionType.CRYSTALLIZE
    none = ElementalReactionType.NONE
    assert check_elemental_reaction(p, [c, d]) == (pc, [p, c], [d])
    assert check_elemental_reaction(c, [c, d]) == (none, [], [c, d])
    assert check_elemental_reaction(c, [d]) == (none, [], [c, d])
    assert check_elemental_reaction(c, [h]) == (ch, [h, c], [])
    assert check_elemental_reaction(c, []) == (none, [], [c])
    assert check_elemental_reaction(c, [p]) == (pc, [p, c], [])
    assert check_elemental_reaction(c, [e]) == (ec, [e, c], [])
    assert check_elemental_reaction(p, [e]) == (pe, [p, e], [])
    assert check_elemental_reaction(e, [p]) == (pe, [p, e], [])
    assert check_elemental_reaction(e, [c, d]) == (ec, [e, c], [d])
    assert check_elemental_reaction(d, [c]) == (none, [], [c, d])
    assert check_elemental_reaction(g, [p]) == (gg, [g, p], [])
    assert check_elemental_reaction(g, [h]) == (gg, [g, h], [])
    assert check_elemental_reaction(g, [e]) == (gg, [g, e], [])
    assert check_elemental_reaction(g, [c]) == (gg, [g, c], [])
    assert check_elemental_reaction(g, [d]) == (none, [], [d])
    assert check_elemental_reaction(g, []) == (none, [], [])
    assert check_elemental_reaction(a, [p]) == (aa, [a, p], [])
    assert check_elemental_reaction(a, [h]) == (aa, [a, h], [])
    assert check_elemental_reaction(a, [e]) == (aa, [a, e], [])
    assert check_elemental_reaction(a, [c]) == (aa, [a, c], [])
    assert check_elemental_reaction(a, [d]) == (none, [], [d])
    assert check_elemental_reaction(a, []) == (none, [], [])


if __name__ == "__main__":
    # test_burning_flame()
    # test_crystallize()
    # test_dendro_core_catalyzing_field()
    # test_frozen()
    # test_swirl_with_catalyzing_field()
    # test_pure_elemental_reaction()
    # test_swirl_with_catalyzing_field_and_dendro_core()
    # test_background_overloaded()
    test_catalyzing_field_old()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AOESummonBase

from ...event import RemoveObjectEventArguments

from ...action import Actions, CreateObjectAction, RemoveObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    TalentBase,
)


# Summons


class DarkfireFurnace_3_7(AOESummonBase):
    name: Literal["Darkfire Furnace"] = "Darkfire Furnace"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1
    back_damage: int = 1


# Skills


class OminousStar(ElementalBurstBase):
    name: Literal["Ominous Star"] = "Ominous Star"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon("Darkfire Furnace")]


class FieryRebirth(CreateStatusPassiveSkill):
    name: Literal["Fiery Rebirth"] = "Fiery Rebirth"
    status_name: Literal["Fiery Rebirth"] = "Fiery Rebirth"
    regenerate_when_revive: bool = False


# Talents


class EmbersRekindled_3_7(TalentBase):
    name: Literal["Embers Rekindled"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal[
        "Abyss Lector: Fathomless Flames"
    ] = "Abyss Lector: Fathomless Flames"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=2)
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].characters:
            if c.name == "Abyss Lector: Fathomless Flames":
                ret.append(c.position)
        return ret

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def _attach_status(self, match) -> List[Actions]:
        """
        If self character do not have Fiery Rebirth, remove self and attach
        Aegis of Abyssal Flame to character.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        status = character.status
        fiery_rebirth_found = False
        for s in status:
            if s.name == "Fiery Rebirth":
                fiery_rebirth_found = True
                break
        if not fiery_rebirth_found:
            # remove self and attach Aegis of Abyssal Flame to character
            return [
                RemoveObjectAction(
                    object_position=self.position,
                ),
                CreateObjectAction(
                    object_position=self.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_name="Aegis of Abyssal Flame",
                    object_arguments={},
                ),
            ]
        return []

    def equip(self, match: Any) -> List[Actions]:
        return self._attach_status(match)

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.CHARACTER:
            return []
        return self._attach_status(match)


# character base


class AbyssLectorFathomlessFlames_3_7(CharacterBase):
    name: Literal["Abyss Lector: Fathomless Flames"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.PYRO
    max_hp: int = 6
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | ElementalSkillBase | OminousStar | FieryRebirth
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Flame of Salvation",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Searing Precept",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            OminousStar(),
            FieryRebirth(),
        ]


register_class(
    AbyssLectorFathomlessFlames_3_7 | DarkfireFurnace_3_7 | EmbersRekindled_3_7
)



from typing import Any, List, Literal, Tuple

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageValue
from ...event import RoundPrepareEventArguments

from ...action import Actions, CreateObjectAction, MakeDamageAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    PlayerActionLabels,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    TalentBase,
)


# Summons


class ChainsOfWardingThunder_3_7(AttackerSummonBase):
    name: Literal["Chains of Warding Thunder"] = "Chains of Warding Thunder"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    increase_cost_valid: bool = True

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When reaching prepare, reset usage
        """
        self.increase_cost_valid = True
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        When increase_cost_valid, enemy perform character switch will increase
        1 cost.
        """
        if not self.increase_cost_valid:
            # out of usage, not modify
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER == 0:
            # not switch character, not modify
            return value
        if value.position.player_idx == self.position.player_idx:
            # self switch, not modify
            return value
        # add 1 cost
        value.cost.any_dice_number += 1
        if mode == "REAL":
            self.increase_cost_valid = False
        return value


# Skills


class RockPaperScissorsCombo(ElementalSkillBase):
    name: Literal["Rock-Paper-Scissors Combo"] = "Rock-Paper-Scissors Combo"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=5)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create prepare skill
        """
        return super().get_actions(match) + [
            self.create_character_status("Rock-Paper-Scissors Combo: Scissors")
        ]


class RockPaperScissorsComboScissors(ElementalSkillBase):
    name: Literal[
        "Rock-Paper-Scissors Combo: Scissors"
    ] = "Rock-Paper-Scissors Combo: Scissors"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Cannot use by player directly
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create prepare skill
        """
        return [
            self.attack_opposite_active(
                match,
                self.damage,
                self.damage_type,
            ),
            self.create_character_status("Rock-Paper-Scissors Combo: Paper"),
        ]


class RockPaperScissorsComboPaper(ElementalSkillBase):
    name: Literal[
        "Rock-Paper-Scissors Combo: Paper"
    ] = "Rock-Paper-Scissors Combo: Paper"
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Cannot use by player directly
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Do not charge.
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]


class LightningLockdown(ElementalBurstBase):
    name: Literal["Lightning Lockdown"] = "Lightning Lockdown"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon("Chains of Warding Thunder")
        ]


class ElectroCrystalCore(CreateStatusPassiveSkill):
    name: Literal["Electro Crystal Core"] = "Electro Crystal Core"
    status_name: Literal["Electro Crystal Core"] = "Electro Crystal Core"
    regenerate_when_revive: bool = False


# Talents


class AbsorbingPrism_3_7(TalentBase):
    name: Literal["Absorbing Prism"]
    version: Literal["3.7"]
    character_name: Literal["Electro Hypostasis"] = "Electro Hypostasis"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EVENT.value
    )
    remove_when_used: bool = True

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return PlayerActionLabels.CARD.value, True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MakeDamageAction | CreateObjectAction]:
        """
        Using this card will heal electro hypostasis by 3 hp and attach
        electro crystal core to it. No need to equip it.
        """
        character = match.player_tables[self.position.player_idx].get_active_character()
        assert character.name == self.character_name
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-3,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            ),
            CreateObjectAction(
                object_name="Electro Crystal Core",
                object_position=character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            ),
        ]


class AbsorbingPrism_4_2(AbsorbingPrism_3_7):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=2)


# character base


class ElectroHypostasis_3_7(CharacterBase):
    name: Literal["Electro Hypostasis"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | RockPaperScissorsCombo
        | RockPaperScissorsComboScissors
        | RockPaperScissorsComboPaper
        | LightningLockdown
        | ElectroCrystalCore
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Electro Crystal Projection",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            RockPaperScissorsCombo(),
            RockPaperScissorsComboScissors(),
            RockPaperScissorsComboPaper(),
            LightningLockdown(),
            ElectroCrystalCore(),
        ]


register_class(
    ElectroHypostasis_3_7
    | AbsorbingPrism_3_7
    | AbsorbingPrism_4_2
    | ChainsOfWardingThunder_3_7
)



from typing import Any, List, Literal

from ....utils.class_registry import register_base_class

from ...struct import Cost

from ...modifiable_values import (
    DamageDecreaseValue,
    DamageElementEnhanceValue,
    DamageValue,
)
from ..base import StatusBase
from ...consts import (
    ELEMENT_TO_ENCHANT_ICON,
    DamageElementalType,
    DamageType,
    ElementType,
    IconType,
    ObjectType,
    SkillType,
)
from ...action import MakeDamageAction, RemoveObjectAction, Actions
from ...event import (
    ChooseCharacterEventArguments,
    MakeDamageEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
    SwitchCharacterEventArguments,
)


class TeamStatusBase(StatusBase):
    """
    Base class of team status.
    """

    type: Literal[ObjectType.TEAM_STATUS] = ObjectType.TEAM_STATUS
    name: str
    version: str
    usage: int
    max_usage: int


register_base_class(TeamStatusBase)


class UsageTeamStatus(TeamStatusBase):
    """
    Base class of team status that based on usages.
    It will implement a method to check if run out of usage; when run out
    of usage, it will remove itself.

    By default, it listens to MAKE_DAMAGE event to check if run out of usage,
    as most usage-based status will decrease usage when damage made.
    Not using AFTER_MAKE_DAMAGE because we should remove the status before
    side effects of elemental reaction, which is triggered on RECEIVE_DAMAGE.

    If it is not a damage related status, check_should_remove should be
    called manually.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when usage has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [
                RemoveObjectAction(
                    object_position=self.position,
                )
            ]
        return []

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When damage made, check whether the team status should be removed.
        """
        return self.check_should_remove()


class RoundTeamStatus(TeamStatusBase):
    """
    Base class of team status that based on rounds.  It will implement the
    event trigger on ROUND_PREPARE to check if run out of usages; when run out
    of usages, it will remove itself.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the status should be removed.
        when round has changed, call this function to check if the status
        should be removed.
        """
        if self.usage <= 0:
            return [
                RemoveObjectAction(
                    object_position=self.position,
                )
            ]
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When reaching prepare, decrease usage and check whether the team
        status should be removed.
        """
        self.usage -= 1
        return list(self.check_should_remove())


class DefendTeamStatus(UsageTeamStatus):
    """
    Base class of defend status (purple shield), decrease damage with its rule
    when receive damage, and decrease usage by 1.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    min_damage_to_trigger: int
    max_in_one_time: int
    decrease_usage_by_damage: bool = False

    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        """
        Decrease damage with its rule, and decrease 1 usage.
        """
        if not value.is_corresponding_character_receive_damage(
            self.position,
            match,
        ):
            # not this character receive damage, not modify
            return value
        new_usage = value.apply_shield(
            self.usage,
            self.min_damage_to_trigger,
            self.max_in_one_time,
            self.decrease_usage_by_damage,
        )
        assert mode == "REAL"
        self.usage = new_usage
        return value


class ShieldTeamStatus(DefendTeamStatus):
    """
    Base class of shield status (yellow shield), decrease damage by its usage
    and decrease corresponding usage.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    icon_type: Literal[IconType.SHIELD] = IconType.SHIELD

    # papams passing to apply_shield is fixed
    min_damage_to_trigger: int = 0
    max_in_one_time: int = 0
    decrease_usage_by_damage: bool = True


class ExtraAttackTeamStatus(TeamStatusBase):
    name: str
    version: str

    trigger_skill_type: SkillType | None
    damage: int
    damage_elemental_type: DamageElementalType
    decrease_usage: bool

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        if skill used by self player, and is trigger skill type or trigger
        skill type is none, then make damage to opponent player's active
        character.
        If self.decrease_usage is True, when make damage success,
        decrease usage by 1.
        """
        assert self.usage >= 0
        if self.position.player_idx != event.action.position.player_idx:
            # not self player use skill
            return []
        if (
            self.trigger_skill_type is not None
            and event.action.skill_type != self.trigger_skill_type
        ):
            # not trigger skill type
            return []
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        if self.decrease_usage:
            self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ]
            )
        ]


class ElementalInfusionTeamStatus(TeamStatusBase):
    """
    Base class of doing elemental infusion. It will infuse all physical damage
    to corresponding elemental damage for active character. If there extra
    conditions, filter the condition and call value modifier in subclasses.
    """

    name: Literal[
        "Pyro Elemental Infusion",
        "Hydro Elemental Infusion",
        "Electro Elemental Infusion",
        "Cryo Elemental Infusion",
        "Anemo Elemental Infusion",
        "Geo Elemental Infusion",
        "Dendro Elemental Infusion",
    ]
    infused_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL

    icon_type: Literal[
        IconType.ELEMENT_ENCHANT_ELEC,
        IconType.ELEMENT_ENCHANT_FIRE,
        IconType.ELEMENT_ENCHANT_WATER,
        IconType.ELEMENT_ENCHANT_ICE,
        IconType.ELEMENT_ENCHANT_WIND,
        IconType.ELEMENT_ENCHANT_ROCK,
        IconType.ELEMENT_ENCHANT_GRASS,
        IconType.ATK_UP,
    ] = IconType.ATK_UP

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.infused_elemental_type == DamageElementalType.PHYSICAL:
            # not set elemental type manually, get it from name
            element = self.name.split(" ")[0].upper()
            assert element in [
                "PYRO",
                "HYDRO",
                "ELECTRO",
                "CRYO",
                "ANEMO",
                "GEO",
                "DENDRO",
            ], "In ElementalInfusion: element not set right value"
            assert self.name.split(" ")[1:] == ["Elemental", "Infusion"]
            self.infused_elemental_type = DamageElementalType(element)
        if self.icon_type == IconType.ATK_UP:
            self.icon_type = ELEMENT_TO_ENCHANT_ICON[
                ElementType(self.infused_elemental_type.value)
            ]  # type: ignore

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        When self use skill, change physical to corresponding element.
        NOTE: this function will not change the usage.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character use skill, do nothing
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return value
        # change physical to other element
        if value.damage_elemental_type == DamageElementalType.PHYSICAL:
            value.damage_elemental_type = self.infused_elemental_type
        return value


class SwitchActionTeamStatus(UsageTeamStatus):
    name: str
    version: str
    usage: int
    max_usage: int

    def _action(self, match: Any) -> List[Actions]:
        """
        action to perform when switch character
        """
        raise NotImplementedError()

    def _check(self, match: Any, event_player_idx: int) -> List[Actions]:
        if event_player_idx != self.position.player_idx:
            # not self switch character
            return []
        if self.usage <= 0:  # pragma: no cover
            # no usage
            return []
        self.usage -= 1
        return self._action(match)

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self switch character, perform attack
        """
        return self._check(match, event.action.player_idx)

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self choose character (when ally defeated), perform attack
        """
        return self._check(match, event.action.player_idx)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)
# Summons


class DrunkenMist_3_3(AttackerSummonBase):
    name: Literal["Drunken Mist"] = "Drunken Mist"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=our_active.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        )
        return ret


# Skills


class IcyPaws(ElementalSkillBase):
    name: Literal["Icy Paws"] = "Icy Paws"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {}
        if self.is_talent_equipped(match):
            # have talent, shield gain one more usage
            args["usage"] = 2
            args["max_usage"] = 2
        return super().get_actions(match) + [
            self.create_team_status("Cat-Claw Shield", args)
        ]


class SignatureMix(ElementalBurstBase):
    name: Literal["Signature Mix"] = "Signature Mix"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        our_active = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=our_active.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        )
        ret.append(self.create_summon("Drunken Mist"))
        return ret


# Talents


class ShakenNotPurred_4_1(SkillTalent):
    name: Literal["Shaken, Not Purred"]
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Diona"] = "Diona"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    skill: Literal["Icy Paws"] = "Icy Paws"


class ShakenNotPurred_3_3(ShakenNotPurred_4_1):
    version: Literal["3.3"]
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=4)


# character base


class Diona_3_3(CharacterBase):
    name: Literal["Diona"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | IcyPaws | SignatureMix] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Kätzlein Style",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            IcyPaws(),
            SignatureMix(),
        ]


register_class(Diona_3_3 | ShakenNotPurred_3_3 | ShakenNotPurred_4_1 | DrunkenMist_3_3)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    check_name,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_barbara():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
            "TEST 1 9 10 10 10 9 10",
            "TEST 2 only p0c0 hydro",
            "sw_char 1 0",
            "end",
            "TEST 3 sw_char 1 cost",
            "skill 1 0 1 2",
            "TEST 3 sw_char 1 cost",
            "sw_char 0 0",
            "sw_char 1 0",
            "end",
            "TEST 3 sw_char 1 cost",
            "skill 1 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 4 summon loop2 oz2",
            "end",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 1 0 1 2",
            "end",
            "TEST 1 7 10 10 0 10 10",
            "TEST 5 team status 1 field",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 1 0 1 2",
            "end",
            "end",
            "TEST 1 9 10 10 9 8 9",
            "TEST 3 sw_char 1 cost",
            "card 0 0 0 1 2 3",
            "TEST 3 sw_char 0 cost",
            "sw_char 2",
            "TEST 3 sw_char 1 cost",
            "skill 2 0 1 2 3 4",
            "end",
            "TEST 1 5 9 10 10 8 10",
            "TEST 2 p0c1 p1c1 p1c2 hydro",
            "TEST 3 sw_char 0 cost",
            "sw_char 1",
            "skill 0 0 1 2",
            "TEST 3 sw_char 1 cost",
            "sw_char 0 0",
            "end",
            "TEST 3 sw_char 1 cost",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 3 sw_char 0 cost",
            "sw_char 0",
            "end",
            "choose 1",
            "TEST 1 7 10 10 0 6 9",
            "skill 2 0 1 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Barbara
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        # Adeptus' Temptation*2
        # Lotus Flower Crisp*2
        # Mondstadt Hash Brown*2
        # Tandoori Roast Chicken
        Glorious Season*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 1:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                parts = cmd.strip().split()
                found_part = set()
                for part in parts:
                    if len(part) == 4 and part[0] == "p" and part[2] == "c":
                        found_part.add(part)
                for player_idx in range(2):
                    for character_idx in range(3):
                        character = match.player_tables[player_idx].characters[
                            character_idx
                        ]
                        part = f"p{player_idx}c{character_idx}"
                        if part in found_part:
                            assert character.element_application == ["HYDRO"]
                        else:
                            assert character.element_application == []
            elif test_id == 3:
                cost = int(cmd.strip().split()[3])
                for req in match.requests:
                    if req.name == "SwitchCharacterRequest":
                        assert req.cost.any_dice_number == cost
            elif test_id == 4:
                summons = match.player_tables[0].summons
                assert len(summons) == 2
                check_name("Melody Loop", summons)
                check_name("Oz", summons)
                for summon in summons:
                    assert summon.usage == 2
            elif test_id == 5:
                team_status = match.player_tables[0].team_status
                assert len(team_status) == 1
                assert team_status[0].name == "Catalyzing Field"
                assert team_status[0].usage == 1
            else:
                break
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_barbara_2():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "end",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 2 0 1 2",
            "TEST 1 6 10 10 7 7 1",
            "skill 1 0 1 2",
            "end",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 1 4 10 10 10 9 0",
            "sw_char 1 0",
            "card 0 0 0 1 2 3",
            "TEST 3 sw_char 0 cost",
            "sw_char 0",
            "TEST 3 sw_char 1 cost",
            "sw_char 1 0",
        ],
        [
            "sw_card",
            "choose 1",
            "card 0 0 0 1 2 3",
            "TEST 3 sw_char 0 cost",
            "sw_char 2",
            "card 0 0 0 1 2 3",
            "TEST 3 sw_char 0 cost",
            "TEST 2 1 summon",
            "end",
            "TEST 3 sw_char 0 cost",
            "TEST 1 8 10 10 9 9 5",
            "sw_char 0",
            "TEST 3 sw_char 0 cost",
            "sw_char 2",
            "TEST 3 sw_char 1 cost",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "choose 1",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "end",
            "TEST 3 sw_char 0 cost",
            "sw_char 0",
            "TEST 3 sw_char 1 cost",
            "sw_char 1 0",
            "TEST 1 4 10 10 8 5 0",
            "skill 2 0 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Barbara*2
        Glorious Season*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = [int(x) for x in cmd.strip().split()[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                summons = match.player_tables[1].summons
                assert len(summons) == 1
                assert summons[0].name == "Melody Loop"
                assert summons[0].usage == 2
            elif test_id == 3:
                cost = int(cmd.strip().split()[3])
                for req in match.requests:
                    if req.name == "SwitchCharacterRequest":
                        assert req.cost.any_dice_number == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_barbara()
    test_barbara_2()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...event import RoundEndEventArguments

from ...action import Actions, ChangeObjectUsageAction, ChargeAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class AfterSalesServiceRounds_4_2(AttackerSummonBase):
    name: Literal["After-Sales Service Rounds"] = "After-Sales Service Rounds"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1


class Jinni_4_2(AttackerSummonBase):
    name: Literal["Jinni"] = "Jinni"
    version: Literal["4.2"] = "4.2"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL
    damage: int = -2
    talent_activated: bool = False

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChargeAction | ChangeObjectUsageAction]:
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        charge = 1
        if self.talent_activated:
            if active_character.hp <= 6:
                self.damage = -3
            if active_character.charge == 0:
                charge = 2
        ret = super().event_handler_ROUND_END(event, match) + [
            ChargeAction(
                player_idx=self.position.player_idx,
                character_idx=active_character.position.character_idx,
                charge=charge,
            )
        ]
        self.damage = -2
        return ret


# Skills


class SpiritWaridingLampTroubleshooterCannon(ElementalSkillBase):
    name: Literal[
        "Spirit-Warding Lamp: Troubleshooter Cannon"
    ] = "Spirit-Warding Lamp: Troubleshooter Cannon"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon("After-Sales Service Rounds")
        ]


class AlcazarzaraysExactitude(ElementalBurstBase):
    name: Literal["Alcazarzaray's Exactitude"] = "Alcazarzaray's Exactitude"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon(
                "Jinni", {"talent_activated": self.is_talent_equipped(match)}
            )
        ]


# Talents


class DiscretionarySupplement_4_2(SkillTalent):
    name: Literal["Discretionary Supplement"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Dori"] = "Dori"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )
    skill: Literal["Alcazarzaray's Exactitude"] = "Alcazarzaray's Exactitude"


# character base


class Dori_4_2(CharacterBase):
    name: Literal["Dori"]
    version: Literal["4.2"] = "4.2"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | SpiritWaridingLampTroubleshooterCannon
        | AlcazarzaraysExactitude
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Marvelous Sword-Dance (Modified)",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpiritWaridingLampTroubleshooterCannon(),
            AlcazarzaraysExactitude(),
        ]


register_class(
    Dori_4_2 | DiscretionarySupplement_4_2 | Jinni_4_2 | AfterSalesServiceRounds_4_2
)



from typing import Dict, List, Literal, Set, Tuple

from .....utils.class_registry import register_class
from ....modifiable_values import DamageIncreaseValue
from ....status.character_status.base import CharacterStatusBase
from ....action import Actions, CreateDiceAction, CreateObjectAction, RemoveObjectAction
from ....event import AfterMakeDamageEventArguments, ReceiveDamageEventArguments
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....character.character_base import (
    CharacterBase,
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    TalentBase,
)
from ....consts import (
    DAMAGE_TYPE_TO_ELEMENT,
    ELEMENT_TO_DAMAGE_TYPE,
    ELEMENT_TO_DIE_COLOR,
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    ElementalReactionType,
    FactionType,
    IconType,
    ObjectPositionType,
    PlayerActionLabels,
    SkillType,
    WeaponType,
)
from .....utils.desc_registry import DescDictType


class StoneFacetsElementalCrystallization_4_3(CharacterStatusBase):
    """
    No actions will be taken by this status; all related actions are handled
    in Stone Facets: Elemental Absorption.
    """

    name: Literal[
        "Stone Facets: Elemental Crystallization",
    ] = "Stone Facets: Elemental Crystallization"
    version: Literal["4.3"] = "4.3"
    usage: int = 0
    max_usage: int = 0
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


class StoneFacetsElementalAbsorption_4_3(CharacterStatusBase):
    """
    It will handle all side effects of Azhdaha's skills, which includes:
    1. When Azhdaha makes elemental skill damage and triggers crystallize,
        it will absorb the element.
    2. When Azhdaha makes elemental skill damage but not trigger crystallize,
        it will generate character status Stone Facets: Elemental
        Crystallization.
    3. When Azhdaha takes elemental damage, and has character status Stone
        Facets: Elemental Crystallization, it will absorb the element.
    4. When Azhdaha absorbs an element, it will override current element
        recorded in the status, and generate elemental die if needed.
    5. When Azhdaha uses elemental burst, based on the number of elements
        absorbed, it will increase damage.
    """

    name: Literal[
        "Stone Facets: Elemental Absorption",
    ] = "Stone Facets: Elemental Absorption"
    version: Literal["4.3"] = "4.3"
    usage: int = 0
    max_usage: int = 0
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    current_damage_element: DamageElementalType = DamageElementalType.GEO
    absorbed_elements: List[DamageElementalType] = []

    def absorb_element(self, element: ElementType) -> List[CreateDiceAction]:
        """
        Handle 4.
        """
        dmg_element = ELEMENT_TO_DAMAGE_TYPE[element]
        res: List[CreateDiceAction] = []
        if dmg_element != self.current_damage_element:
            # generate die
            res.append(
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    color=ELEMENT_TO_DIE_COLOR[element],
                )
            )
        self.current_damage_element = dmg_element
        if dmg_element not in self.absorbed_elements:
            self.absorbed_elements.append(dmg_element)
            self.usage = len(self.absorbed_elements)
        return res

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        """
        Handle 1, 2 and 3.
        """
        dmg = event.final_damage
        if (
            self.position.check_position_valid(
                dmg.position,
                match,
                player_idx_same=True,
                character_idx_same=True,
                target_area=ObjectPositionType.SKILL,
            )
            and not dmg.damage_from_element_reaction
        ):
            # is self use skill damage, check for 1 and 2
            skill = match.get_object(dmg.position)
            assert skill is not None
            if skill.name == "Aura of Majesty":
                # is geo elemental skill
                if dmg.element_reaction == ElementalReactionType.CRYSTALLIZE:
                    # trigger crystallize, 1
                    target_element = dmg.reacted_elements[1]
                    assert target_element != ElementType.GEO
                    return list(self.absorb_element(target_element))
                else:
                    # not trigger crystallize, 2
                    return [
                        CreateObjectAction(
                            object_name=("Stone Facets: Elemental Crystallization"),
                            object_position=self.position,
                            object_arguments={},
                        )
                    ]
        elif self.position.check_position_valid(
            dmg.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
        ):
            # is self take damage, check for 3
            if (
                dmg.damage_type != DamageType.DAMAGE
                or dmg.damage_elemental_type
                not in [
                    DamageElementalType.CRYO,
                    DamageElementalType.HYDRO,
                    DamageElementalType.PYRO,
                    DamageElementalType.ELECTRO,
                ]
                or dmg.damage_elemental_type == self.current_damage_element
            ):
                # not damage, or elemental type not right, or currently
                # absorbed this element, do nothing
                return []
            crystallization_status = None
            for s in (
                match.player_tables[self.position.player_idx]
                .characters[self.position.character_idx]
                .status
            ):
                if s.name == "Stone Facets: Elemental Crystallization":
                    crystallization_status = s
                    break
            if crystallization_status is None:
                # not have crystallization status, do nothing
                return []
            # absorb element, and remove crystallization status
            return list(
                self.absorb_element(DAMAGE_TYPE_TO_ELEMENT[dmg.damage_elemental_type])
                + [RemoveObjectAction(object_position=crystallization_status.position)]
            )
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["REAL", "TEST"]
    ) -> DamageIncreaseValue:
        """
        Handle 5.
        """
        if (
            not value.is_corresponding_character_use_damage_skill(
                self.position, match, SkillType.ELEMENTAL_BURST
            )
            or value.damage_from_element_reaction
        ):
            # not corresponding character use elemental burst,
            # or from elemental reaction, do nothing
            return value
        assert mode == "REAL"
        # increase damage
        value.damage += len(self.absorbed_elements)
        return value


class AuraOfMajesty(ElementalSkillBase):
    """
    In this skill, only make damage. All side effects will be handled in
    Stone Facets: Elemental Absorption.
    """

    name: Literal["Aura of Majesty"] = "Aura of Majesty"
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=3,
    )


class AzhdahaOtherElementalSkill(ElementalSkillBase):
    """
    For Azhdaha, it can use other elemental skills when absorbed an element.
    This class do normal damage, except that will check whether Azhdaha has
    absorbed an element, so that can use it.
    """

    name: Literal[
        "Frostspike Wave",
        "Torrential Rebuke",
        "Blazing Rebuke",
        "Thunderstorm Wave",
    ]
    # these two attributes will be override in __init__
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost()

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        if self.name == "Frostspike Wave":
            self.damage_type = DamageElementalType.CRYO
        elif self.name == "Torrential Rebuke":
            self.damage_type = DamageElementalType.HYDRO
        elif self.name == "Blazing Rebuke":
            self.damage_type = DamageElementalType.PYRO
        elif self.name == "Thunderstorm Wave":
            self.damage_type = DamageElementalType.ELECTRO
        else:
            raise AssertionError("Unknown skill name")
        self.cost = Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[
                DAMAGE_TYPE_TO_ELEMENT[self.damage_type]
            ],
            elemental_dice_number=3,
        )

    def is_valid(self, match: Match) -> bool:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        status = character.status
        for s in status:
            if s.name == "Stone Facets: Elemental Absorption":  # pragma: no branch
                # find the status, check whether it has absorbed right element
                return (
                    s.current_damage_element  # type: ignore
                    == self.damage_type
                )
        raise AssertionError("Stone Facets: Elemental Absorption not found")

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_character_status("Stone Facets: Elemental Crystallization")
        ]


class DecimatingRockfall(ElementalBurstBase):
    """
    Its damage increase is handled in Stone Facets: Elemental Absorption.
    """

    name: Literal["Decimating Rockfall"] = "Decimating Rockfall"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=2
    )


class StoneFacets(CreateStatusPassiveSkill):
    name: Literal["Stone Facets"] = "Stone Facets"
    status_name: Literal[
        "Stone Facets: Elemental Absorption"
    ] = "Stone Facets: Elemental Absorption"


class LunarCyclesUnending_4_3(TalentBase):
    name: Literal["Lunar Cycles Unending"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Azhdaha"] = "Azhdaha"
    cost: Cost = Cost(same_dice_number=2)
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EVENT.value
    )
    remove_when_used: bool = True

    def get_action_type(self, match: Match) -> Tuple[int, bool]:
        return PlayerActionLabels.CARD.value, True

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateDiceAction | CreateObjectAction]:
        """
        Create status, and create dice.
        """
        characters = match.player_tables[self.position.player_idx].characters
        character = match.player_tables[self.position.player_idx].get_active_character()
        assert character.name == self.character_name
        elements: Set[ElementType] = set()
        for c in characters:
            elements.add(c.element)
        ret: List[CreateDiceAction | CreateObjectAction] = [
            CreateObjectAction(
                object_name="Stone Facets: Elemental Crystallization",
                object_position=character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]
        for e in elements:
            ret.append(
                CreateDiceAction(
                    player_idx=self.position.player_idx,
                    number=1,
                    color=ELEMENT_TO_DIE_COLOR[e],
                )
            )
        return ret


class Azhdaha_4_3(CharacterBase):
    name: Literal["Azhdaha"]
    desc: Literal["", "CRYO", "ELECTRO", "PYRO", "HYDRO"] = ""
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | AuraOfMajesty
        | AzhdahaOtherElementalSkill
        | DecimatingRockfall
        | StoneFacets
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Sundering Charge",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            AuraOfMajesty(),
            AzhdahaOtherElementalSkill(name="Frostspike Wave"),
            AzhdahaOtherElementalSkill(name="Torrential Rebuke"),
            AzhdahaOtherElementalSkill(name="Blazing Rebuke"),
            AzhdahaOtherElementalSkill(name="Thunderstorm Wave"),
            DecimatingRockfall(),
            StoneFacets(),
        ]

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Match
    ) -> List[Actions]:
        """
        After make damage, based on element absorbed, change self desc.
        """
        if self.is_defeated:  # pragma: no cover
            return []
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        target: StoneFacetsElementalAbsorption_4_3 | None = None
        for s in status:  # pragma: no branch
            if s.name == "Stone Facets: Elemental Absorption":  # pragma: no branch
                target = s  # type: ignore
                break
        assert target is not None
        element = target.current_damage_element
        if element == DamageElementalType.GEO:
            self.desc = ""
        elif element == DamageElementalType.CRYO:
            self.desc = "CRYO"
        elif element == DamageElementalType.ELECTRO:
            self.desc = "ELECTRO"
        elif element == DamageElementalType.PYRO:
            self.desc = "PYRO"
        elif element == DamageElementalType.HYDRO:
            self.desc = "HYDRO"
        else:
            raise AssertionError(f"Unknown element {element.name}")
        return []


desc: Dict[str, DescDictType] = {
    "CHARACTER/Azhdaha": {
        "names": {"en-US": "Azhdaha", "zh-CN": "若陀龙王"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_Dahaka.png",  # noqa: E501
        "id": 2602,
    },
    "CHARACTER/Azhdaha_CRYO": {
        "names": {"en-US": "Azhdaha·Cryo", "zh-CN": "若陀龙王·冰"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_DahakaIce.png",  # noqa: E501
    },
    "CHARACTER/Azhdaha_ELECTRO": {
        "names": {"en-US": "Azhdaha·Electro", "zh-CN": "若陀龙王·雷"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_DahakaElec.png",  # noqa: E501
    },
    "CHARACTER/Azhdaha_PYRO": {
        "names": {"en-US": "Azhdaha·Pyro", "zh-CN": "若陀龙王·火"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_DahakaFire.png",  # noqa: E501
    },
    "CHARACTER/Azhdaha_HYDRO": {
        "names": {"en-US": "Azhdaha·Hydro", "zh-CN": "若陀龙王·水"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_DahakaWater.png",  # noqa: E501
    },
    "SKILL_Azhdaha_NORMAL_ATTACK/Sundering Charge": {
        "names": {"en-US": "Sundering Charge", "zh-CN": "碎岩冲撞"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Azhdaha_ELEMENTAL_SKILL/Frostspike Wave": {
        "names": {"en-US": "Frostspike Wave", "zh-CN": "霜刺破袭"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Cryo DMG, and attach Stone Facets: Elemental Crystallization to this character.",  # noqa: E501
                "zh-CN": "造成3点冰元素伤害，此角色附属磐岩百相·元素凝晶。",
            }
        },
    },
    "SKILL_Azhdaha_ELEMENTAL_SKILL/Torrential Rebuke": {
        "names": {"en-US": "Torrential Rebuke", "zh-CN": "洪流重斥"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Hydro DMG, and attach Stone Facets: Elemental Crystallization to this character.",  # noqa: E501
                "zh-CN": "造成3点水元素伤害，此角色附属磐岩百相·元素凝晶。",
            }
        },
    },
    "SKILL_Azhdaha_ELEMENTAL_SKILL/Blazing Rebuke": {
        "names": {"en-US": "Blazing Rebuke", "zh-CN": "炽焰重斥"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Pyro DMG, and attach Stone Facets: Elemental Crystallization to this character.",  # noqa: E501
                "zh-CN": "造成3点火元素伤害，此角色附属磐岩百相·元素凝晶。",
            }
        },
    },
    "SKILL_Azhdaha_ELEMENTAL_SKILL/Thunderstorm Wave": {
        "names": {"en-US": "Thunderstorm Wave", "zh-CN": "霆雷破袭"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Electro DMG, and attach Stone Facets: Elemental Crystallization to this character.",  # noqa: E501
                "zh-CN": "造成3点雷元素伤害，此角色附属磐岩百相·元素凝晶。",
            }
        },
    },
    "SKILL_Azhdaha_ELEMENTAL_SKILL/Aura of Majesty": {
        "names": {"en-US": "Aura of Majesty", "zh-CN": "磅礴之气"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Geo DMG. If a Crystallize reaction occurs, then this character will perform Elemental Absorption.\nIf during the usage of this skill, the character hasn't absorbed an element's power yet, then Stone Facets: Elemental Crystallization will be attached.",  # noqa: E501
                "zh-CN": "造成3点岩元素伤害，如果发生了结晶反应，则角色汲取对应元素的力量。\n如果本技能中角色未汲取元素的力量，则附属磐岩百相·元素凝晶。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Stone Facets: Elemental Crystallization": {
        "names": {
            "en-US": "Stone Facets: Elemental Crystallization",
            "zh-CN": "磐岩百相·元素凝晶",
        },
        "descs": {
            "4.3": {
                "en-US": "After this character takes Cryo/Hydro/Pyro/Electro DMG: If this character currently has not absorbed the power of this element, remove this effect, and then this character performs Elemental Absorption.",  # noqa: E501
                "zh-CN": "角色受到冰/水/火/雷元素伤害后：如果角色当前未汲取该元素的力量，则移除此状态，然后角色汲取对应元素的力量。",  # noqa: E501
            }
        },
        "image_path": "status/Dahaka_S.png",
    },
    "SKILL_Azhdaha_ELEMENTAL_BURST/Decimating Rockfall": {
        "names": {"en-US": "Decimating Rockfall", "zh-CN": "山崩毁阵"},
        "descs": {
            "4.3": {
                "en-US": "Deals 4 Geo DMG. DMG +1 for each element previously absorbed.",  # noqa: E501
                "zh-CN": "造成4点岩元素伤害，每汲取过一种元素此伤害+1。",
            }
        },
    },
    "SKILL_Azhdaha_PASSIVE/Stone Facets": {
        "names": {"en-US": "Stone Facets", "zh-CN": "磐岩百相"},
        "descs": {
            "4.3": {
                "en-US": "(Passive) When the battle begins, this character gains Stone Facets: Elemental Absorption.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，初始附属磐岩百相·元素汲取。",
            }
        },
    },
    "CHARACTER_STATUS/Stone Facets: Elemental Absorption": {
        "names": {
            "en-US": "Stone Facets: Elemental Absorption",
            "zh-CN": "磐岩百相·元素汲取",
        },
        "descs": {
            "4.3": {
                "en-US": "The character can absorb the power of the Elements, including Cryo/Hydro/Pyro/Electro, and then based on the element absorbed, gain the skills Frostspike Wave/Torrential Rebuke/Blazing Rebuke/Thunderstorm Wave. (The character can only have one Elemental Type absorbed at once, and this status will record the Elemental Type previously absorbed)\nAfter this character absorbs an element of a different type than the currently absorbed element: Generate 1 Elemental Die of the same type as the Element that was just absorbed. The usages of this status represents the number of Elemental Types that the character has absorbed.",  # noqa: E501
                "zh-CN": "角色可以汲取冰/水/火/雷元素的力量，然后根据所汲取的元素类型，获得技能霜刺破袭/洪流重斥/炽焰重斥/霆雷破袭。（角色同时只能汲取一种元素，此状态会记录角色已汲取过的元素类型数量）\n角色汲取了一种和当前不同的元素后：生成1个所汲取元素类型的元素骰。该角色状态的可用次数代表角色汲取过的元素种类数。",  # noqa: E501
            }
        },
    },
    "TALENT_Azhdaha/Lunar Cycles Unending": {
        "names": {"en-US": "Lunar Cycles Unending", "zh-CN": "晦朔千引"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Azhdaha, play this to them. Attach Stone Facets: Elemental Crystallization to Azhdaha, then create 1 Elemental Die for each Elemental Type your characters have.\n(Deck must contain Azhdaha to add this card to it)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为若陀龙王时，对该角色打出。使若陀龙王附属磐岩百相·元素凝晶，然后生成每种我方角色所具有的元素类型的元素骰各1个。\n（牌组中包含若陀龙王，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Dahaka.png",  # noqa: E501
        "id": 226022,
    },
}


register_class(
    Azhdaha_4_3
    | StoneFacetsElementalAbsorption_4_3
    | StoneFacetsElementalCrystallization_4_3
    | LunarCyclesUnending_4_3,
    desc,
)



from re import Match
from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import UseCardValue
from ....event import RoundPrepareEventArguments, UseCardEventArguments
from ....status.team_status.base import UsageTeamStatus
from ....consts import IconType, ObjectPositionType, ObjectType
from ....action import ConsumeArcaneLegendAction, CreateObjectAction, RemoveObjectAction
from ....struct import Cost, ObjectPosition
from ....card.event.arcane_legend import ArcaneLegendBase
from .....utils.desc_registry import DescDictType


class PassingOfJudgmentStatus_4_3(UsageTeamStatus):
    name: Literal["Passing of Judgment"]
    version: Literal["4.3"] = "4.3"
    usage: int = 3
    max_usage: int = 3
    icon_type: IconType = IconType.DEBUFF

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        In prepare, remove self
        """
        self.usage = 0
        return self.check_should_remove()

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        check should remove
        """
        return self.check_should_remove()

    def value_modifier_USE_CARD(
        self, value: UseCardValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> UseCardValue:
        """
        When self use a event card, make using card failed.
        """
        if value.position.player_idx != self.position.player_idx:
            # not self, return
            return value
        if value.card.type in [ObjectType.CARD, ObjectType.ARCANE]:
            # is target card type
            assert self.usage > 0
            assert mode == "REAL"
            if value.use_card:  # pragma: no branch
                value.use_card = False
                self.usage -= 1
        return value


class PassingOfJudgment_4_3(ArcaneLegendBase):
    """
    TODO: will it block talent+event card, e.g. Electro Hypostasis talent?
    """

    name: Literal["Passing of Judgment"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1, arcane_legend=True)

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        return super().get_actions(target, match) + [
            CreateObjectAction(
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_name=self.name,
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Passing of Judgment": {
        "names": {"en-US": "Passing of Judgment", "zh-CN": "裁定之时"},
        "descs": {
            "4.3": {
                "en-US": "The next 3 Event Cards played by you in this Round will take no effect.",  # noqa: E501
                "zh-CN": "本回合中，你接下来打出的3张事件牌无效。",
            }
        },
    },
    "ARCANE/Passing of Judgment": {
        "names": {"en-US": "Passing of Judgment", "zh-CN": "裁定之时"},
        "descs": {
            "4.3": {
                "en-US": '3 Event Cards played by your opponent in this Round will take no effect.\n(Only one "Arcane Legend" card can be played per match. This card will be in your starting hand.)',  # noqa: E501
                "zh-CN": "本回合中，对方牌手打出的3张事件牌无效。\n（整局游戏只能打出一张「秘传」卡牌；这张牌一定在你的起始手牌中）",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Xuanpanyouzui.png",  # noqa: E501
        "id": 330006,
    },
}


register_class(PassingOfJudgment_4_3 | PassingOfJudgmentStatus_4_3, desc)
